"""Place ARD Orders

Provides
--------
  1. Place an ARD order
  2. Test an ARD order with "dry-run" mode
  3. Check order status

Reference
---------

Ordering API overview

- https://ard.maxar.com/docs/select-and-order/order-ard/

SDK Ordering Tutorial

- https://ard.maxar.com/sdk-docs/sdk/ordering-ard/

API request and response examples

- https://ard.maxar.com/docs/api-reference/post-order/
- https://ard.maxar.com/docs/api-reference/get-order-status/


Example
-------

>>> from max_ard import Order
>>> order = Order(destination='my_bucket/my_ard_tiles', select_id=<Select ID>)
>>> order.add_email_notification('me@email.com')
>>> order.submit()

"""




from typing import Union, List
from functools import wraps
from posixpath import join as urljoin
import warnings

from pydantic import BaseModel
from max_ard.session import get_user_session
from max_ard.exceptions import ARDServerException, BadARDRequest, NotSubmitted

from max_ard.session import ARD_HOSTNAME, ARD_DEV_BRANCH
from max_ard.admin import UsageArea, UsageCost

__all__ = ('Order',)

POST_ORDER_URL = f'https://{ARD_HOSTNAME}/order{ARD_DEV_BRANCH}'
GET_ORDERS_URL = f'https://{ARD_HOSTNAME}/order{ARD_DEV_BRANCH}'
GET_ORDER_URL = f'https://{ARD_HOSTNAME}/order/status{ARD_DEV_BRANCH}'



#######
# Models
#######

class Acquisition(BaseModel):
    id: str
    cells: List[str] = []

    class Config:
        validate_assignment = True


class Output_Config(BaseModel):
    bucket: str 
    prefix: str 
    role_arn: str = None

    class Config:
        validate_assignment = True


class EmailNotification(BaseModel):
    type: str = 'email'
    address: str
 

    class Config:
        allow_mutation: False

class SNSNotification(BaseModel):
    type: str = 'sns'
    topic_arn: str
        
    class Config:
        allow_mutation: False


class OrderRequest(BaseModel):
    acquisitions: List[Acquisition] = None
    select_id: str = None
    dry_run: bool = False
    # filter
    intersects: str = None
    bbox:  List = None
    # destination
    output_config: Output_Config = None
    bundle_adjust: bool = False
    notifications: List[Union[EmailNotification, SNSNotification]] = []
    metadata: dict = {}

    class Config:
        extra = 'allow'
        validate_assignment = True

    def to_payload(self):
        ''' sends only parameters with values '''
        params = self.dict().copy()
        payload = {k:v for k,v in params.items() if v not in (None, [], ())}
        params = payload['output_config'].copy()
        payload['output_config'] = {k:v for k,v in params.items() if v is not None}
        return payload


class OrderUsage(BaseModel):
    area: UsageArea = None
    cost: UsageCost = None

class OrderDetails(BaseModel):
    usage: OrderUsage = None

    class Config:
        extra = 'allow'

class OrderResponse(BaseModel):
    id: str
    status: str
    status_message: str
    order: OrderDetails = None

    class Config:
        extra = 'allow'
        validate_assignment = True


#####
# Classes
#####


class Order():
    ''' An ARD API Order object

    Parameters
    ----------
    acquisitions : iterable of str or dict, optional
        An iterable of acquisitions to order, see Notes
    select_id : str 
        A Select ID to order, see Notes
    destination: 
        Path to storage location, such as s3://my-bucket/my-prefix
    intersects : Shapely geometry or str, optional
        Only generate tiles that intersect this geometry, can be Shapely or WKT
    bbox : interable of numeric, optional
        Like `intersects`, a bounding box in WGS84 coordinates, [XMIN YMIN XMAX YMAX]
    role_arn : str, optional
        A trusted Role ARN for the writer to assume so it can write tiles. 
        This is not used if the s3 bucket policy allows writing tiles to the bucket.
    dry_run : bool, optional
        When true, runs pre-order checks to check if order is valid but does not generate imagery
    bba : bool, optional
        When true, Block Bundle Adjustment will be applied to the order.
    metadata : dict, optional 
        User-supplied metadata 

    Attributes
    ----------
    dry_run : bool
        as above
    submitted : bool
        True if Select has been submitted via `max_ard.order.Order.submit`
    request : Pydantic model
        Parameters are loaded into a Pydantic model of the HTTP API request
    response : Pydantic model
        Pydantic model of the server response to API call. Status and submit calls
        return the same payload.

    Notes
    -----
    An order must specify `acquisitions` or `select_id`.

    If ordering by select ID, do not include acquisition IDs or an AOI in the order request.

    Acquisitions can be a list of acqusitions IDs, or a dict where acquisition IDs are keys
    and their values are lists of ARD grid cell IDs. '''

    def __init__(self,
        acquisitions = None,
        select_id = None,
        destination = None,
        intersects = None,
        bbox = None,
        role_arn = None,
        dry_run = False,
        bba = False,
        metadata = {},
        ):
        
    
        # set up the output config
        if destination is not None:
            destination = destination.replace('s3://', '')
            bucket, prefix = destination.split('/', 1)
            output_config = Output_Config(bucket=bucket, prefix=prefix, role_arn=role_arn)
        else:
            output_config = None
        

        # set up acquisitions
        if type(acquisitions) == str:
            acquisitions = [Acquisition(id=acquisitions)]
        elif type(acquisitions) == list or type(acquisitions) == tuple:
            if len(acquisitions) > 0:
                if type(acquisitions[0]) == str:
                    acquisitions = [Acquisition(id=v) for v in acquisitions]
        elif type(acquisitions) == dict:
            acquisitions = [Acquisition(id=k, cells=v) for k,v in acquisitions.items()]

        # if the intersects is a Shapely geom, convert it to wkt
        try:
            intersects = intersects.wkt
        except:
            pass

        self.request = OrderRequest(
            acquisitions=acquisitions,
            select_id=select_id,
            intersects=intersects,
            bbox=bbox,
            output_config=output_config,
            bundle_adjust=bba,
            dry_run=dry_run,
            metadata=metadata
        )
        self.dry_run = dry_run
        self.response = None
        self.submitted = False


    class _validators():
    
        @staticmethod
        def needs_submittal(func):
            @wraps(func)
            def func_wrapper(instance, *args, **kwargs):
                if not instance.submitted:
                    raise NotSubmitted(f"Can't access {func.__name__} until Order has been submitted")
                res = func(instance, *args, **kwargs)
                return res
            return func_wrapper


    def add_email_notification(self, address):
        ''' Add an email notification to the order

        Parameters
        ----------
        address : str
            Email address to receive order notifications '''
        self.request.notifications.append(EmailNotification(address=address))

    def add_sns_notification(self, topic_arn):
        ''' Add an AWS SNS notification topic to receive order notifications

        Parameters
        ----------
        topic_arn : str
            AWS SNS topic ARN to recieve order notifications '''
        self.request.notifications.append(SNSNotification(topic_arn=topic_arn))


    def __repr__(self):
        try:
            return f'<Order {self.order_id} ({self.status})>'
        except NotSubmitted:
            return '<Order (Not submitted)>'

    @property
    @_validators.needs_submittal
    def finished(self):
        ''' The Order finished processing but may have failed'''

        return self.status != 'RUNNING'

    @property
    @_validators.needs_submittal
    def running(self):
        ''' The Order is running'''

        return self.status == 'RUNNING'

    @property
    @_validators.needs_submittal
    def succeeded(self):
        ''' The Order has finished running and has succeeded '''

        return self.status == 'SUCCEEDED'

    @property
    @_validators.needs_submittal
    def usage(self):
        if self.response.order is None:
            # just submitted, refresh the order status
            response = self.get_order(self.order_id)
            self.response = OrderResponse(**response)
        return self.response.order.usage

    @property
    @_validators.needs_submittal
    def failed(self):
        ''' The Order has finished running but failed'''

        return self.status in ['FAILED', 'ERROR']

    @property
    @_validators.needs_submittal
    def order_id(self):
        ''' The Order ID'''

        return self.response.id

    @property
    @_validators.needs_submittal
    def status(self):
        ''' State of the order process: 'RUNNING', 'SUCCEEDED, or 'FAILED' '''

        if self.response.status == 'RUNNING':
            response = self.get_order(self.order_id)
            self.response = OrderResponse(**response) 
        return self.response.status

    @property
    def state(self):
        ''' Legacy version of `status`, will be deprecated in the future '''
        return self.status


    @classmethod
    def from_id(cls, order_id, session=None):
        ''' Create a Order object from an ID

        Parameters
        ----------
        order_id: str
            Order ID to hydrate into a Order object
        session : Session object, optional 
            Authenticated session, such as from get_client_session()

        Returns
        -------
        Order ''' 

        response = cls.get_order(order_id, session)
        instance = cls()
        instance.response = OrderResponse(**response) 
        # can recreate the request
        instance.submitted = True
        return instance

    @classmethod
    def list_orders(cls, limit=10, starting_after='', ending_before='', filter='', session=None):
        ''' Fetch user's orders

        Parameters
        ----------
        limit: int, default 10
            maximum number of orders to fetch
        starting_after: str
            the order_id after which further responses will be returned, paging forward
        ending_before: str
            the order_id before which further responses will be returned, paging backwards
        filter: str
            filter results that match values contained in the given key separated by a colon. 
            Example: 'metadata.downstream_customer_id:abdc-534-b4dc47'
        session: Session object, optional
            Authenticated session, such as from get_client_session()

        Returns
        -------
        list
            Order objects matching parameters ''' 

        if not session:
            session = get_user_session()

        params = {
            'limit': limit,
            'starting_after': starting_after,
            'ending_before': ending_before,
            'filter': filter
        }

        response = session.get(GET_ORDERS_URL, params=params)
        if response.status_code == 400:
            raise BadARDRequest(response.json()['message'])
        # server error
        elif response.status_code != 200:
            raise ARDServerException(response.json()['message'])
        # success
        orders = []
        for order in response.json()['data']:
            instance = cls()
            #TODO remove faked status fields
            instance.response = OrderResponse(**order, status='SUCCEEDED', status_message = '') 
            instance.submitted = True
            orders.append(instance)
        return orders

    @classmethod
    def get_order(cls, order_id, session=None):
        ''' Fetch raw data about an Order from an ID

        Parameters
        ----------
        order_id : str
            Order ID to fetch metadata for
        session: Session object, optional
            Authenticated session, such as from get_client_session()

        Returns
        -------
        dict
            API data for the given Order ''' 

        if not session:
            session = get_user_session()

        url = urljoin(GET_ORDER_URL, order_id)
        r = session.get(url)
        if r.status_code == 400:
            raise BadARDRequest(r.json()['message'])
        # server error
        elif r.status_code != 200:
            raise ARDServerException(r.json()['message'])
        # success
        return r.json()


    @classmethod
    def send_order(cls, payload, session=None):
        ''' Send a request to the Order API

        Parameters
        ----------
        payload : dict
            Order API request payload
        session : Session object, optional
            Authenticated session, such as from get_client_session()

        Returns
        -------
        dict
            API response data for the given Order '''

        if not session:
            session = get_user_session()
        r = session.post(POST_ORDER_URL, json=payload)
        # bad format
        if r.status_code == 400:
            raise BadARDRequest(r.json()['message'])
        # server error
        elif r.status_code != 202:
            raise ARDServerException(r.json()['message'])
        # success
        return r.json()
    

    def submit(self):
        ''' Submit this Order to the API '''

        response = self.send_order(self.request.to_payload())
        if self.submitted:
            warnings.warn('The Order has already been submitted')
            return
        self.submitted = True
        if self.dry_run:
            self.response = OrderResponse(
                id='dry run',
                status='SUCCEEDED',
                order={'usage': response['usage']},
                status_message='Order dry-run validation successful.'
            )
        else:
            self.response = OrderResponse(
                id=response['id'],
                status='RUNNING',
                status_message='Order submitted and running.'
            )