"""Tools and utilities for processing ARD images

Provides
--------
- read/write windows from ARD collections
"""
from max_ard.exceptions import MissingDependency
from shapely.geometry import shape
from shapely.ops import transform
from shapely.geometry.base import BaseGeometry

from maxar_canvas_grid.grid import get_transform, get_CRS
from maxar_canvas_grid import covers

from max_ard.dependency_support import HAS_RASTERIO
if HAS_RASTERIO:
    import rasterio
    from rasterio import windows

        
__all__ = ('read_windows', 'write_windows')

def prep_geoms(geoms):
    ''' Prepare geometry input(s) for looping 

        Args:
            geoms: one or more geometry objects as:
                - shapely geometry objects
                - dicts that follow the Python Geometry Interface (GeoJSON-like)
                - GeoJSON feature-like dicts that have a 'geometry' key
                - lists or tuples of above
                - a Fiona dataset reader

        Returns:
            list: shapely geometry objects of input(s)'''
    
    # shapely geometry, json-like (__geo_interface__), or fiona collection
    # iterable
    try:
        return [shape(geom) for geom in geoms]
    except:
        pass

    # single feature
    try:
        return [shape(geoms)]
    except:
        pass

    # geojson or fiona collection
    # iterable
    try:
        return [shape(geom['geometry']) for geom in geoms]
    except:
        pass
        
    # single feature
    try:
        return [shape(geoms['geometry'])]
    except:
        pass

    # if none of the type conditions have been satisfied, throw an exception
    raise TypeError("Invalid argument--check list of acceptable types.")


def loop_geoms(collection, geoms, src_proj=4326):

    src_crs = get_CRS(src_proj)

    for geom in prep_geoms(geoms):
            for cell in covers(geom, src_proj=src_proj):
                for tile in collection.get_stack(cell):
                    tile_crs = get_CRS(tile.cell.proj)
                    tfm = get_transform(src_crs, tile_crs)
                    proj_geom = transform(tfm, geom)
                    yield tile, proj_geom

                

def read_windows(collection, geoms, src_proj=4326):
    ''' A generator for reading windows from overlapping tiles in an ARDCollection

            Args:
                geoms: one or more geometry objects as:
                    - shapely geometry objects
                    - dicts that follow the Python Geometry Interface (GeoJSON-like)
                    - GeoJSON feature-like dicts that have a 'geometry' key
                    - lists or tuples of above
                    - a Fiona dataset reader
                src_proj: An EPSG identifier parseable by Proj4

            Yields:
                Tuple of:
                    ARDTile of tile
                    Shapely polygon representing the bounds of the input geometry
                        in the tile's UTM coordinate system
                    a Reader function:

                        reader(asset)

                            Args:
                                asset(str): asset to read: `visual`, `pan`, or `ms`
                            
                            Returns:
                                numpy array of data covered by the returned geometry

            Example:

            geoms = [...]
            for tile, geom, reader in read_windows(geoms):
                # can check tile.properties, or fetch tile.clouds
                # can check if `geom` intersects with clouds
                data = reader('visual') '''

    if not HAS_RASTERIO:
        raise MissingDependency('Rasterio is required to read ARD rasters')

    for tile, geom in loop_geoms(collection, geoms, src_proj):

        def reader(asset, **kwargs):
            dsr = tile.open_asset(asset)
            window = rasterio.windows.from_bounds(*geom.bounds, dsr.transform)
            return dsr.read(window=window, **kwargs)

        yield (tile, geom, reader)


def write_windows(collection, geoms, src_proj=4326):

    if not HAS_RASTERIO:
        raise MissingDependency('Rasterio is required to read ARD rasters')

    for tile, geom in loop_geoms(collection, geoms, src_proj):
    
        def writer(asset, name, **kwargs):
            dsr = tile.open_asset(asset)
            window = rasterio.windows.from_bounds(*geom.bounds, dsr.transform)
            window = windows.from_bounds(*geom.bounds, dsr.transform)
            profile = dsr.meta
            profile.update({
                'width': window.width,
                'height': window.height,
                'count': dsr.count,
                'transform': dsr.window_transform(window),
                'driver': 'GTiff'
            })
            profile.update(kwargs)
            with rasterio.open(name, 'w', **profile) as dest:
                dest.write(dsr.read(window=window))

        yield (tile, geom, writer)