''' Authenticated sessions for communicating with the ARD API endpoints'''

import os
import json
from configparser import ConfigParser
import jwt
from jwt.api_jwt import DecodeError
from time import time
import sys
import warnings

from oauthlib.oauth2 import LegacyApplicationClient, BackendApplicationClient
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2.rfc6749.errors import MissingTokenError
from requests.auth import HTTPBasicAuth

__all__ = ('get_user_session', 'get_self')

SAVE_TOKEN = True # if false, never save the token back to the config file
ARD_HOSTNAME = os.environ.get('MAXAR_ARD_HOSTNAME', 'ard.maxar.com/api/v1')
ARD_DEV_BRANCH = os.environ.get('ARD_DEV_BRANCH', None)

if ARD_DEV_BRANCH and not ARD_DEV_BRANCH.startswith('-'):
    ARD_DEV_BRANCH = '-' + ARD_DEV_BRANCH
else:
    ARD_DEV_BRANCH = ''

AUTH_URL = f'https://{ARD_HOSTNAME}/auth{ARD_DEV_BRANCH}/authenticate'


_USER_SESSION_CACHE = None

def get_self(session):
    """ Fetch the account SELF endpoint of the current user

    Parameters
    ----------
    session : OAuth2Session
        The session object created for the current user

    Returns
    -------
    dict: SELF api payload """

    r = session.get(f'https://{ARD_HOSTNAME}/auth{ARD_DEV_BRANCH}/self')
    r.raise_for_status()
    return r.json()

def get_session():
    #warnings.warn('get_session() will be deprecated for the more specific get_user_session()')
    return get_user_session()

def get_user_session():
    """ Get an authenticated Requests session for communicating with the ARD API

    Example
    -------
    >>> r = get_user_session()
    >>> resp = r.get('https://ard-dev.geobigdata.io/select/request/5542092796671610498')

    Returns
    -------
    OAuth2Session
        Authenticated Requests session

    Notes
    -----
    For credentials, create a file in your home directory called .ard-config with these contents:

        [ard]
        user_name = <your_user_name>
        user_password = <your_password>

    or set these environment variables:

        ARD_USERNAME
        ARD_PASSWORD
    """
    global _USER_SESSION_CACHE
    if _USER_SESSION_CACHE is None:
        _USER_SESSION_CACHE = _get_user_session()
    return _USER_SESSION_CACHE

def _get_user_session(config_file=None):
    """ Get an authenticated Requests session for communicating with the ARD API

    Parameters
    ----------
    config_file, optional
        Path to config file

    Returns
    -------
    OAuth2Session
        Authenticated Requests session

    Notes
    -----
    If you provide ARD_ACCESS_TOKEN and ARD_REFRESH_TOKEN via env vars it will
    use those credentials.  If you provide a path to a config
    file, it will look there for the credentials. If you don't it will try to
    pull the credentials from environment variables (ARD_USERNAME, ARD_PASSWORD).  
    If that fails and you have a '~/.ard-config' ini file, it will read from that.
    """

    if os.environ.get("ARD_ACCESS_TOKEN", None):
        return session_from_existing_token(access_token=os.environ.get("ARD_ACCESS_TOKEN", None), refresh_token=os.environ.get("ARD_REFRESH_TOKEN", None))

    # If no config file specified, try using environment variables.  If that
    # fails and there is a config in the default location, use that.
    if not config_file:
        try:
            return session_from_envvars()
        except MissingTokenError as e:
            raise Exception('Invalid ARD credentials given in environment variables.')
        except Exception as e:
            config_file = os.path.expanduser('~/.ard-config')

    example_credentials = """
    [ard]
    user_name = your_user_name
    user_password = your_password
    """

    if not os.path.isfile(config_file):
        raise Exception(f"Please create a ARD credential file at ~/.ard-config with these contents:\n{example_credentials}")

    try:
      session = session_from_config(config_file)
    except:
      raise Exception("Invalid credentials or incorrectly formated config file at ~/.ard-config")

    return session

def session_from_existing_token(access_token, refresh_token="no_refresh_token", auth_url=AUTH_URL):
    """Get a session using an existing access_token and refresh_token.

    Parameters
    ----------
    access_token : str
        Access token
    refresh_token : str
        Refresh token

    Returns
    -------
    OAuth2Session
        Authenticated Requests session
    """
    def save_token(token):
        s.token = token

    token = {"token_type": "Bearer", "refresh_token": refresh_token, "access_token": access_token, "scope": ["read", "write"], "expires_in": 604800, "expires_at": 0}

    # grab the expiration from the jwt access_token. Don't verify because
    # we dont care about source of token. ARD auth endpoint will handle that.
    try:
        t = jwt.decode(access_token, verify=False)
    except DecodeError:
        raise Exception('Supplied ARD access token is not a valid JWT.  Check ARD_ACCESS_TOKEN env var or runtime.json')
    token['expires_at'] = t['exp']

    s = OAuth2Session(
            token=token,
            auto_refresh_url=auth_url,
            token_updater=save_token,
            )
    return s

def session_from_envvars(auth_url=AUTH_URL,
                         environ_template=(('username', 'ARD_USERNAME'),
                                           ('password', 'ARD_PASSWORD'))):
    """Returns a session with the ARD authorization token baked in,
    pulling the credentials from environment variables.

    environ_template - An iterable of key, value pairs. The key should
                      be the variables used in the oauth workflow, and
                      the values being the environment variables to
                      pull the configuration from.  Change the
                      template values if your envvars differ from the
                      default, but make sure the keys remain the same.
    Returns
    -------
    OAuth2Session
        Authenticated Requests session
    """
    def save_token(token):
        s.token = token

    client_id = 'dummyclientid'
    client_secret = 'dummyclientsecret'

    environ = {var:os.environ[envvar] for var, envvar in environ_template}
    s = OAuth2Session(client=LegacyApplicationClient(client_id),
                      auto_refresh_url=auth_url,
                      auto_refresh_kwargs={'client_id':client_id,
                                           'client_secret':client_secret},
                      token_updater=save_token)

    s.fetch_token(auth_url, **environ)

    return s

def session_from_kwargs(**kwargs):
    """Get a  session object using credentials from keywords
    
    Parameters
    ----------
    **kwargs : dict  
        Authentication keywords: username, password, client_id, client_secret

    Returns
    -------
    OAuth2Session
        Authenticated Requests session
    """
    def save_token(token):
        s.token = token
    auth_url=AUTH_URL
    s = OAuth2Session(client=LegacyApplicationClient(kwargs.get('client_id')),
                      auto_refresh_url=auth_url,
                      auto_refresh_kwargs={'client_id':kwargs.get('client_id'),
                                           'client_secret':kwargs.get('client_secret')},
                      token_updater=save_token)

    try:
        s.fetch_token(auth_url,
                  username=kwargs.get('username'),
                  password=kwargs.get('password'),
                  client_id=kwargs.get('client_id'),
                  client_secret=kwargs.get('client_secret'))
    except MissingTokenError as e:
        raise Exception('Invalid credentials passed into session_from_kwargs()')

    return s



def session_from_config(config_file):
    """Get a  session object using credentials from a config file
    
    Parameters
    ----------
    config_file : str
        path to config file

    Returns
    -------
    OAuth2Session
        Authenticated Requests session
    """

    def save_token(token_to_save):
        """Save off the token back to the config file."""
        if not SAVE_TOKEN:
            return
        if not 'ard_token' in set(cfg.sections()):
            cfg.add_section('ard_token')
        cfg.set('ard_token', 'json', json.dumps(token_to_save))
        with open(config_file, 'w') as sink:
            cfg.write(sink)

    # Read the config file (ini format).
    cfg = ConfigParser(interpolation=None)
    if not cfg.read(config_file):
        raise RuntimeError('No ini file found at {} to parse.'.format(config_file))
    client_id = 'dummy_client_id(not-required)'
    client_secret = 'dummy_client_secret(not-required)'

    # the ini file has the optional ability to set an auth url (useful for dev)
    if not cfg.has_option('ard','auth_url'):
        auth_url = AUTH_URL
    else:
        auth_url = cfg.get('ard','auth_url')

    # See if we have a token stored in the config, and if not, get one.
    if 'ard_token' in set(cfg.sections()):
        # Parse the token from the config.
        token = json.loads(cfg.get('ard_token','json'))

        # the refresh token expires 18 days after the auth token
        refresh_expires = token['expires_at'] + (18 * 24 * 3600)

        # pad the time by 30 seconds
        if time() + 30 < refresh_expires:

            s = OAuth2Session(client_id, client=LegacyApplicationClient(client_id, token=token),
                            auto_refresh_url=auth_url,
                            auto_refresh_kwargs={'client_id':client_id,
                                                'client_secret':client_secret},
                            token_updater=save_token)
            s.token = token

            # hit the server to trigger a refresh if the token has expired
            if time() + 30 < token['expires_at']:
                s.get(f'https://{ARD_HOSTNAME}/auth{ARD_DEV_BRANCH}/self')
            return s

    # No pre-existing token or the refresh expired, so we request one from the API.
    s = OAuth2Session(client_id, client=LegacyApplicationClient(client_id),
                        auto_refresh_url=auth_url,
                        auto_refresh_kwargs={'client_id':client_id,
                                            'client_secret':client_secret},
                        token_updater=save_token)

    # Get the token and save it to the config.
    token = s.fetch_token(auth_url,
                            username=cfg.get('ard','user_name'),
                            password=cfg.get('ard','user_password'),
                            client_id=client_id,
                            client_secret=client_secret)
    save_token(token)

    return s

def write_config(username, password):
    """Write a config file

    The config file will be written to ~/.ard-config.

    Parameters
    ----------
    username : str
        Username
    password: str
        Password
    """

    config_file = os.path.expanduser('~/.ard-config')
    cfg = ConfigParser()
    cfg.read(config_file)

    # create or update ard section
    if 'ard' not in cfg.sections():
        cfg['ard'] = {}
    if username:
        cfg['ard']['user_name'] = username
    if password:
        cfg['ard']['user_password'] = password
    
    # reset the token
    cfg.remove_section('ard_token') 

    with open(config_file, 'w') as configfile:
        cfg.write(configfile)

def read_token():
    """ Read the token from the config file

    Returns
    -------
    str
        Access token 
    """

    config_file = os.path.expanduser('~/.ard-config')
    cfg = ConfigParser()
    cfg.read(config_file)
    print(json.loads(cfg['ard_token']['json'])['access_token'])
           
